@@ -47,7 +47,10 @@ class TestBase(TestCase):
 
         pattern = ''.join(map(escape, pattern))
         regex = re.sub(r'\s+', r'\\s*', pattern)
-        self.assertRegexpMatches(text, regex)
+        if PY2:
+            self.assertRegexpMatches(text, regex)
+        else:
+            self.assertRegex(text, regex)
 
     def assert_ir_line(self, line, mod):
         lines = [line.strip() for line in str(mod).splitlines()]
@@ -56,6 +59,15 @@ class TestBase(TestCase):
     def assert_valid_ir(self, mod):
         llvm.parse_assembly(str(mod))
 
+    def assert_pickle_correctly(self, irobject):
+        """Assert that the IR object pickles and unpickles correctly.
+        The IR string is equal and that their type is equal
+        """
+        newobject = pickle.loads(pickle.dumps(irobject, protocol=-1))
+        self.assertIs(irobject.__class__, newobject.__class__)
+        self.assertEqual(str(irobject), str(newobject))
+        return newobject
+
     def module(self):
         return ir.Module()
 
@@ -129,6 +141,8 @@ class TestFunction(TestBase):
         self.assertEqual(asm,
             ("declare %s alwaysinline optsize alignstack(16) "
              "personality i8 (...)* @\"__gxx_personality_v0\"") % self.proto)
+        # Check pickling
+        self.assert_pickle_correctly(func)
 
     def test_function_attributes(self):
         # Now with parameter attributes
@@ -143,6 +157,8 @@ class TestFunction(TestBase):
         self.assertEqual(asm,
             """declare noalias i32 @"my_func"(i32 zeroext %".1", i32 dereferenceable(5) dereferenceable_or_null(10) %".2", double %".3", i32* nonnull align 4 %".4")"""
             )
+        # Check pickling
+        self.assert_pickle_correctly(func)
 
     def test_function_metadata(self):
         # Now with function metadata
@@ -153,6 +169,8 @@ class TestFunction(TestBase):
         self.assertEqual(asm,
             """declare i32 @"my_func"(i32 %".1", i32 %".2", double %".3", i32* %".4") !dbg !0"""
             )
+        # Check pickling
+        self.assert_pickle_correctly(func)
 
     def test_define(self):
         # A simple definition
@@ -193,6 +211,10 @@ class TestFunction(TestBase):
         powi2 = module.declare_intrinsic('llvm.powi', [dbl])
         self.assertIs(powi, powi2)
 
+    def test_pickling(self):
+        fn = self.function()
+        self.assert_pickle_correctly(fn)
+
 
 class TestIR(TestBase):
 
@@ -466,6 +488,10 @@ class TestGlobalValues(TestBase):
             @"g" = internal global i32 123, align 16
             """)
 
+    def test_pickle(self):
+        mod = self.module()
+        self.assert_pickle_correctly(mod)
+
 
 class TestBlock(TestBase):
 
@@ -489,6 +515,12 @@ class TestBlock(TestBase):
         d = builder.sub(a, b, 'd')
         builder.mul(d, b, 'e')
         f = ir.Instruction(block, a.type, 'sdiv', (c, b), 'f')
+        self.check_block(block, """\
+            my_block:
+                %"c" = add i32 %".1", %".2"
+                %"d" = sub i32 %".1", %".2"
+                %"e" = mul i32 %"d", %".2"
+            """)
         block.replace(d, f)
         self.check_block(block, """\
             my_block:
@@ -637,6 +669,24 @@ class TestBuildInstructions(TestBase):
                 %"d" = xor i32 %".2", -1
             """)
 
+    def test_replace_operand(self):
+        block = self.block(name='my_block')
+        builder = ir.IRBuilder(block)
+        a, b = builder.function.args[:2]
+        undef1 = ir.Constant(ir.IntType(32), ir.Undefined)
+        undef2 = ir.Constant(ir.IntType(32), ir.Undefined)
+        c = builder.add(undef1, undef2, 'c')
+        self.check_block(block, """\
+            my_block:
+                %"c" = add i32 undef, undef
+            """)
+        c.replace_usage(undef1, a)
+        c.replace_usage(undef2, b)
+        self.check_block(block, """\
+            my_block:
+                %"c" = add i32 %".1", %".2"
+            """)
+
     def test_integer_comparisons(self):
         block = self.block(name='my_block')
         builder = ir.IRBuilder(block)
@@ -768,6 +818,11 @@ class TestBuildInstructions(TestBase):
         self.assertEqual(h.type, ir.VoidType())
         i = builder.load(c, 'i', align=1)
         self.assertEqual(i.type, int32)
+        # Atomics
+        j = builder.store_atomic(b, c, ordering="seq_cst", align=4)
+        self.assertEqual(j.type, ir.VoidType())
+        k = builder.load_atomic(c, ordering="seq_cst", align=4, name='k')
+        self.assertEqual(k.type, int32)
         # Not pointer types
         with self.assertRaises(TypeError):
             builder.store(b, a)
@@ -788,6 +843,8 @@ class TestBuildInstructions(TestBase):
                 %"g" = load i32, i32* %"c"
                 store i32 %".2", i32* %"c", align 1
                 %"i" = load i32, i32* %"c", align 1
+                store atomic i32 %".2", i32* %"c" seq_cst, align 4
+                %"k" = load atomic i32, i32* %"c" seq_cst, align 4
             """)
 
     def test_gep(self):
@@ -1165,6 +1222,64 @@ class TestBuildInstructions(TestBase):
                 call void @"llvm.assume"(i1 %"c")
             """)
 
+    def test_vector_ops(self):
+        block = self.block(name='insert_block')
+        builder = ir.IRBuilder(block)
+        a, b = builder.function.args[:2]
+        a.name = 'a'
+        b.name = 'b'
+
+        vecty = ir.VectorType(a.type, 2)
+        vec = ir.Constant(vecty, ir.Undefined)
+        idxty = ir.IntType(32)
+        vec = builder.insert_element(vec, a, idxty(0), name='vec1')
+        vec = builder.insert_element(vec, b, idxty(1), name='vec2')
+
+        self.check_block(block, """\
+            insert_block:
+                %"vec1" = insertelement <2 x i32> <i32 undef, i32 undef>, i32 %"a", i32 0
+                %"vec2" = insertelement <2 x i32> %"vec1", i32 %"b", i32 1
+            """)
+
+        block = builder.append_basic_block("shuffle_block")
+        builder.branch(block)
+        builder.position_at_end(block)
+
+        mask = ir.Constant(vecty, [1, 0])
+        shuff = builder.shuffle_vector(vec, vec, mask, name='shuf')
+
+        self.check_block(block, """\
+            shuffle_block:
+                %"shuf" = shufflevector <2 x i32> %"vec2", <2 x i32> %"vec2", <2 x i32> <i32 1, i32 0>
+            """)
+
+        block = builder.append_basic_block("add_block")
+        builder.branch(block)
+        builder.position_at_end(block)
+
+        vadd = builder.add(vec, vec, name='sum')
+
+        self.check_block(block, """\
+            add_block:
+                %"sum" = add <2 x i32> %"vec2", %"vec2"
+            """)
+
+        block = builder.append_basic_block("extract_block")
+        builder.branch(block)
+        builder.position_at_end(block)
+
+        c = builder.extract_element(vec, idxty(0), name='ex1')
+        d = builder.extract_element(vec, idxty(1), name='ex2')
+
+        self.check_block(block, """\
+            extract_block:
+              %"ex1" = extractelement <2 x i32> %"vec2", i32 0
+              %"ex2" = extractelement <2 x i32> %"vec2", i32 1
+            """)
+
+        builder.ret(builder.add(c, d))
+        self.assert_valid_ir(builder.module)
+
     def test_bitreverse(self):
         block = self.block(name='my_block')
         builder = ir.IRBuilder(block)
@@ -1429,6 +1544,34 @@ class TestBuilderMisc(TestBase):
                 br label %"one.endif"
             """)
 
+
+    def test_if_then_long_label(self):
+        full_label = 'Long'*20
+        block = self.block(name=full_label)
+        builder = ir.IRBuilder(block)
+        z = ir.Constant(int1, 0)
+        a = builder.add(z, z, 'a')
+        with builder.if_then(a) as bbend:
+            b = builder.add(z, z, 'b')
+            with builder.if_then(b) as bbend:
+                c = builder.add(z, z, 'c')
+        builder.ret_void()
+        self.check_func_body(builder.function, """\
+            {full_label}:
+                %"a" = add i1 0, 0
+                br i1 %"a", label %"{label}.if", label %"{label}.endif"
+            {label}.if:
+                %"b" = add i1 0, 0
+                br i1 %"b", label %"{label}.if.if", label %"{label}.if.endif"
+            {label}.endif:
+                ret void
+            {label}.if.if:
+                %"c" = add i1 0, 0
+                br label %"{label}.if.endif"
+            {label}.if.endif:
+                br label %"{label}.endif"
+            """.format(full_label=full_label, label=full_label[:25] + '..'))
+
     def test_if_then_likely(self):
         def check(likely):
             block = self.block(name='one')
@@ -1606,10 +1749,7 @@ class TestTypes(TestBase):
     def test_pickling(self):
         types = self.assorted_types()
         for ty in types:
-            data = pickle.dumps(ty, protocol=-1)
-            newty = pickle.loads(data)
-            self.assertIs(newty.__class__, ty.__class__)
-            self.assertEqual(str(newty), str(ty))
+            newty = self.assert_pickle_correctly(ty)
             if self.has_logical_equality(ty):
                 self.assertEqual(newty, ty)
 
@@ -1753,6 +1893,11 @@ class TestTypes(TestBase):
         td = llvm.create_target_data("e-m:e-i64:64-f80:128-n8:16:32:64-S128")
         self.assertEqual(mytype.get_abi_size(td, context=context), 4)
 
+    def test_vector(self):
+        context = ir.Context()
+        vecty = ir.VectorType(ir.IntType(32), 8)
+        self.assertEqual(str(vecty), "<8 x i32>")
+
 
 c32 = lambda i: ir.Constant(int32, i)
 
@@ -1819,6 +1964,14 @@ class TestConstant(TestBase):
         with self.assertRaises(ValueError):
             ir.Constant(ir.ArrayType(int32, 3), (5, 6))
 
+    def test_vector(self):
+        vecty = ir.VectorType(ir.IntType(32), 8)
+        vals = [1, 2, 4, 3, 8, 6, 9, 7]
+        vec = ir.Constant(vecty, vals)
+        vec_repr = "<8 x i32> <{}>".format(
+            ', '.join(map('i32 {}'.format, vals)))
+        self.assertEqual(str(vec), vec_repr)
+
     def test_structs(self):
         st1 = ir.LiteralStructType((flt, int1))
         st2 = ir.LiteralStructType((int32, st1))
@@ -1849,6 +2002,11 @@ class TestConstant(TestBase):
         with self.assertRaises(ValueError):
             ir.Constant(st2, (4, 5, 6))
 
+    def test_undefined_literal_struct_pickling(self):
+        i8 = ir.IntType(8)
+        st = ir.Constant(ir.LiteralStructType([i8, i8]), ir.Undefined)
+        self.assert_pickle_correctly(st)
+
     def test_type_instantiaton(self):
         """
         Instantiating a type should create a constant.
@@ -1958,5 +2116,13 @@ class TestTransforms(TestBase):
         self.assertEqual(call.callee, bar)
 
 
+class TestSingleton(TestBase):
+    def test_undefined(self):
+        self.assertIs(ir.Undefined, ir.values._Undefined())
+        self.assertIs(ir.Undefined, copy.copy(ir.Undefined))
+        self.assertIs(ir.Undefined, copy.deepcopy(ir.Undefined))
+        self.assert_pickle_correctly(ir.Undefined)
+
+
 if __name__ == '__main__':
     unittest.main()
